# Copyright 2021 Eric Antones <eantones@nuobit.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).
import collections
import logging
import uuid

from odoo import _, fields
from odoo.exceptions import ValidationError

from odoo.addons.component.core import AbstractComponent
from odoo.addons.connector.components.mapper import m2o_to_external

_logger = logging.getLogger(__name__)


class Mapper(AbstractComponent):
    _inherit = "base.mapper"

    def _apply_with_options(self, map_record):
        """
        Hack to allow having non required children field
        """
        assert (
            self.options is not None
        ), "options should be defined with '_mapping_options'"
        _logger.debug("converting record %s to model %s", map_record.source, self.model)

        fields = self.options.fields
        for_create = self.options.for_create
        result = {}
        for from_attr, to_attr in self.direct:
            if isinstance(from_attr, collections.Callable):
                attr_name = self._direct_source_field_name(from_attr)
            else:
                attr_name = from_attr

            if not fields or attr_name in fields:
                value = self._map_direct(map_record.source, from_attr, to_attr)
                result[to_attr] = value

        for meth, definition in self.map_methods:
            mapping_changed_by = definition.changed_by
            if not fields or (
                mapping_changed_by and mapping_changed_by.intersection(fields)
            ):
                if definition.only_create and not for_create:
                    continue
                values = meth(map_record.source)
                if not values:
                    continue
                if not isinstance(values, dict):
                    raise ValueError(
                        "%s: invalid return value for the "
                        "mapping method %s" % (values, meth)
                    )
                result.update(values)

        for from_attr, to_attr, model_name in self.children:
            if not fields or from_attr in fields:
                if from_attr in map_record.source:
                    items = self._map_child(map_record, from_attr, to_attr, model_name)
                    if items:
                        result[to_attr] = items
        return self.finalize(map_record, result)

    def get_target_fields(self, map_record, fields):
        if not fields:
            return []
        fields = set(fields)
        result = {}
        for from_attr, to_attr in self.direct:
            if isinstance(from_attr, collections.Callable):
                # attr_name = self._direct_source_field_name(from_attr)
                # TODO
                raise NotImplementedError
            else:
                if to_attr in fields:
                    if to_attr in result:
                        raise ValidationError(_("Field '%s' mapping defined twice"))
                    result[to_attr] = from_attr

        # TODO: create a new decorator to write the field mapping manually
        #   I think this is not necessary, just use changed_by is precisely for that
        # for meth, definition in self.map_methods:
        #     for mcb in definition.mapping:
        #         if mcb in fields:
        #             if to_attr in result:
        #                 raise ValidationError("Field '%s' mapping defined twice")
        #             result[to_attr] = from_attr

        for from_attr, to_attr, _model_name in self.children:
            if to_attr in fields:
                if to_attr in result:
                    raise ValidationError(_("Field '%s' mapping defined twice"))
                result[to_attr] = from_attr

        return list(set(result.values()))


class ChannelChildMapperImport(AbstractComponent):
    _inherit = "base.map.child"

    def get_all_items(self, mapper, items, parent, to_attr, options):
        mapped = []
        neobookings_user = self.env["res.users"].search(
            [("login", "=", "neobookings@roomdoo.com")]
        )
        neobooking_item_ids = []
        for item in items:
            map_record = mapper.map_record(item, parent=parent)
            if self.skip_item(map_record):
                continue
            item_values = self.get_item_values(map_record, to_attr, options)
            if item_values:
                self._child_bind(map_record, item_values)
                mapped.append(item_values)
                if hasattr(items, "_name"):
                    neobooking_item_ids.append(item.id)
        neobookings_property_ids = neobookings_user.pms_property_ids.ids

        if self.backend_record.pms_property_id.id not in neobookings_property_ids:
            return mapped

        neobookings_property_id = self.backend_record.pms_property_id.id
        neobookings_property = self.env["pms.property"].browse(neobookings_property_id)
        ota_neo_settings = neobookings_property.ota_property_settings_ids.filtered(
            lambda r: "eobookings" in r.agency_id.name
        )
        neobookings_pricelist_id = ota_neo_settings.main_pricelist_id.id
        neobookings_availability_plan_id = ota_neo_settings.main_avail_plan_id.id
        payload = False
        items_to_upload = False
        call_type = False
        min_date = False
        max_date = False
        room_type_ids = False
        if (
            hasattr(items, "_name")
            and items._name == "channel.wubook.product.pricelist.item"
        ):
            call_type = "prices"
            items_to_upload = (
                self.env["channel.wubook.product.pricelist.item"]
                .browse(neobooking_item_ids)
                .filtered(
                    lambda r: r.pricelist_id.id == neobookings_pricelist_id
                    and neobookings_property_id in r.pms_property_ids.ids
                )
            )
            if items_to_upload:
                min_date = min(items_to_upload.mapped("date_end_consumption"))
                max_date = max(items_to_upload.mapped("date_end_consumption"))
                room_type_ids = (
                    self.env["pms.room.type"]
                    .search(
                        [("product_id", "in", items_to_upload.mapped("product_id").ids)]
                    )
                    .ids
                )
                payload, endpoint = neobookings_property.get_payload_prices(
                    prices=items_to_upload, client=neobookings_user
                )
        if hasattr(items, "_name") and items._name == "channel.wubook.pms.availability":
            call_type = "availability"
            items_to_upload = (
                self.env["channel.wubook.pms.availability"]
                .browse(neobooking_item_ids)
                .filtered(lambda r: r.pms_property_id.id == neobookings_property_id)
            )
            if items_to_upload:
                min_date = min(items_to_upload.mapped("date"))
                max_date = max(items_to_upload.mapped("date"))
                room_type_ids = items_to_upload.mapped("room_type_id.id")
                payload, endpoint = neobookings_property.get_payload_avail(
                    avails=items_to_upload, client=neobookings_user
                )
        if (
            hasattr(items, "_name")
            and items._name == "channel.wubook.pms.availability.plan.rule"
        ):
            call_type = "restrictions"
            items_to_upload = (
                self.env["channel.wubook.pms.availability.plan.rule"]
                .browse(neobooking_item_ids)
                .filtered(
                    lambda r: r.availability_plan_id.id
                    == neobookings_availability_plan_id
                    and r.pms_property_id.id == neobookings_property_id
                )
            )
            if items_to_upload:
                min_date = min(items_to_upload.mapped("date"))
                max_date = max(items_to_upload.mapped("date"))
                room_type_ids = items_to_upload.mapped("room_type_id.id")
                payload, endpoint = neobookings_property.get_payload_rules(
                    rules=items_to_upload, client=neobookings_user
                )
        if payload:
            _logger.info("Exporting Neobookings")
            response = neobookings_property.pms_api_push_payload(
                payload=payload, endpoint=endpoint, client=neobookings_user
            )
            self.env["pms.api.log"].sudo().create(
                {
                    "pms_property_id": neobookings_property_id,
                    "client_id": neobookings_user.id,
                    "request": payload,
                    "response": str(response),
                    "status": "success" if response.ok else "error",
                    "request_date": fields.Datetime.now(),
                    "method": "PUSH",
                    "endpoint": endpoint,
                    "target_date_from": min_date,
                    "target_date_to": max_date,
                    "request_type": call_type,
                    "room_type_ids": room_type_ids,
                }
            )
        return mapped

    def get_items(self, items, parent, to_attr, options):
        mapper = self._child_mapper()
        mapped = self.get_all_items(mapper, items, parent, to_attr, options)
        return self.format_items(mapped)

    def _child_bind(self, map_record, item_values):
        return


class ImportMapChildBinder(AbstractComponent):
    _name = "base.map.child.binder.import"
    _inherit = "base.map.child.import"

    def _child_bind(self, map_record, item_values):
        binder = self.binder_for()
        if binder._external_field not in item_values:
            item_values[binder._external_field] = uuid.uuid4().hex
        item_values[binder._sync_date_field] = fields.Datetime.now()


class ExportMapChildBinder(AbstractComponent):
    _name = "base.map.child.binder.export"
    _inherit = "base.map.child.export"

    def _child_bind(self, map_record, item_values):
        binder = self.binder_for()
        external_id = map_record.source.external_id or uuid.uuid4().hex
        binder.bind(external_id, map_record.source, export=True)


# TODO: create a fix on OCA repo and remove this class
class ExportMapper(AbstractComponent):
    _inherit = "base.export.mapper"

    def _map_direct(self, record, from_attr, to_attr):
        """Apply the ``direct`` mappings.

        :param record: record to convert from a source to a target
        :param from_attr: name of the source attribute or a callable
        :type from_attr: callable | str
        :param to_attr: name of the target attribute
        :type to_attr: str
        """
        if isinstance(from_attr, collections.Callable):
            return from_attr(self, record, to_attr)

        value = record[from_attr]
        if value is None:  # we need to allow fields with value 0
            return False

        # Backward compatibility: when a field is a relation, and a modifier is
        # not used, we assume that the relation model is a binding.
        # Use an explicit modifier m2o_to_external  in the 'direct' mappings to
        # change that.
        field = self.model._fields[from_attr]
        if field.type == "many2one":
            mapping_func = m2o_to_external(from_attr)
            value = mapping_func(self, record, to_attr)
        return value


# TODO: move uuid to generic binder
